/**
 * (C) A. Lang 2011
 * Licensed under GPL v3
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#define alloc_blocksize 1024
#define max_lines 512
#include "fnt5x7.h"
#include "fnt8x8.h"

bool omit_margin=false;
int _eps_mode=0;
char _esc_sequence[128];
int _esc_pos=0;
int _substatus=0;
int _line=0;
unsigned char _linebuffer[0x10000];
int _linepos=0;
int _max_linewidth=0x10000;
unsigned char* outbuffer[max_lines];
int linewidth[max_lines];
int _max_used_linewidth;
extern bool use8px;
extern bool singleSpace;

enum _esc_modes {
	esc_text=0,
	esc_seq,
	esc_margin,
	esc_lspacing,
	esc_bitimg

} ;

typedef struct {
	int margin_left;
	int margin_top;
	int linespacing;
} printer_status_t;

unsigned char * esc_buffer=NULL;
printer_status_t _pstat;

int copy_line() {
	if (_line>=max_lines) {
		fprintf(stderr,"Max %i lines of output supported so far. Not writing line %i\n",max_lines,_line);
		return 0;
	}
	if (NULL==(outbuffer[_line]=(unsigned char*)calloc(_linepos,1))) {
		fprintf(stderr,"Could not store line %i\n",_line);
		return 2;
	};
	if (_linepos>_max_used_linewidth)
		_max_used_linewidth=_linepos;
	linewidth[_line]=_linepos;
	memcpy(outbuffer[_line],_linebuffer,_linepos);
	return 0;
}
int newline() {
	if ((_linepos)&&copy_line())
		return 2;
	_linepos=0;
	if (!omit_margin) {
		for (int i=0;i<_pstat.margin_left;i++) {
			_linebuffer[_linepos++]=0;
		}
	}
	_line++;
	return 0;
}
void write_line_data(int length, unsigned char * data, bool flip_data=false) {
	for(int p=0;p<length;p++) {
		if (_linepos>=_max_linewidth) {
			fprintf(stderr,"ERROR: too much data @%i\n",_line);
			return;
		}
		if (flip_data) {
			int tmp=*(data++);
			_linebuffer[_linepos++]=(tmp>>7)|((tmp>>5)&2)|((tmp>>3)&4)|((tmp>>1)&8)|((tmp<<1)&16)|((tmp<<3)&32)|((tmp<<5)&64)|((tmp<<7));
		}else{
			_linebuffer[_linepos++]=*(data++);
		}
	}
}
void write_char(unsigned char c) {
	if (use8px) {
		write_line_data(8, (unsigned char *)&(font8x8int[c*8]), true);
	}else{
		unsigned char sp=0;
		write_line_data(5, (unsigned char *)&(font5x7int[c*5]), true);
		write_line_data(1, &sp, true);
		if (!singleSpace)
			write_line_data(1, &sp, true);
	}
}
void fixme(const char * reason ) {
	if (_esc_pos) {
		fprintf(stderr,"FIXME @<unknown>: %s\n",reason);
		fprintf(stderr,"ESC buffer was: <esc>");
		for (int cp=0; cp<_esc_pos; cp++) {
			char cc=_esc_sequence[cp];
			if (isprint(cc)) {
				fprintf(stderr,"\t%02x",cc & 0xff);
			}else{
				putc(cc,stderr);
			}
		}
		putc('\n',stderr);
	}else{
		fprintf(stderr,"FIXME @<unknown>: %s\n",reason);
	}
}
int img_density, img_ncolumns;

int epson_init() {
	_eps_mode=0;
	_esc_pos=0;
	_substatus=0;
	_line=0;
	_linepos=0;
	_pstat.margin_left=0;
	_pstat.margin_top=0;
	_pstat.linespacing=0;
	_max_used_linewidth=0;
	return 0;
}
int epson_parse(unsigned char c) {
	switch(_eps_mode) {
		case esc_text: //text output
			switch (c) {
				case 27: //start of esc sequence
					_esc_pos=0;
					_eps_mode=1;
					break;
				case 0x0c:
					return 1;
				case 13:
					if (newline()) {
						return -1;
					}
					break;
				case 0xff:
					break;
				default:
					if (c>31) {
						write_char(c);
					}
					break;
			}
			break;
		case esc_margin:
			_pstat.margin_left=c*8;
			_eps_mode=esc_text;
			if ((!omit_margin)&&(_linepos<_pstat.margin_left)) {
				while(_linepos<_pstat.margin_left) {
					_linebuffer[_linepos++]=0;
				}
			}
			break;
		case esc_lspacing:
			_pstat.linespacing=c;
			_eps_mode=esc_text;
			break;
		case esc_bitimg:
			switch (_substatus) {
				case 0:
					img_density=c;
					break;
				case 1:
					img_ncolumns=c;
					break;
				case 2:
					img_ncolumns+=c*256;
					break;
				default:
					write_line_data(1,&c);
					break;
			}
			_substatus++;
			if (_substatus==(3+img_ncolumns)) {
				_eps_mode=esc_text;
				_substatus=0;

			}
			break;
		case esc_seq: // escape sequence read in
			_esc_sequence[_esc_pos++]=c;
			switch (_esc_sequence[0]) {
				case '@': //initialize printer
					epson_init();
					break;
				case 'A':
					_eps_mode=esc_lspacing;
					break;
				case '*':
					_eps_mode=esc_bitimg;
					_substatus=0;
					break;
				case 'l':
					_eps_mode=esc_margin;
					break;
				case '2':
					_eps_mode=esc_text;
					break;
				case '(': //ESC ( command group
					if (_esc_pos==1)
						break;
					switch (_esc_sequence[1]) {
						case 'c': //set page format
								fixme("unknown sequence");
							break;
						case 'C': //set page length in defined unit
								fixme("unknown sequence");
							break;
						default:
							break;
					}
					goto clear_mode;
					break;
				default:
					fixme("unknown_sequence");
					goto clear_mode;
					break;
			}
	}
	return 0;
clear_mode:
	_esc_pos=0;
	_eps_mode=0;
	return 0;
}

