#include <Imlib2.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include "escp.h"
#include <string.h>
#include <getopt.h>
#include <unistd.h>

FILE  * inputfile=NULL;
bool transparent_background=false;
bool invert_colors=false;
bool greenscreen=false;
bool use8px=false, singleSpace=false;

int save_image(char * format, char * filename) {
	Imlib_Image out;
	//fprintf(stderr,"Creating image with %ix%i pixels\n",_max_used_linewidth, 8*_line);
	out = imlib_create_image(_max_used_linewidth, 8*_line);
	imlib_context_set_image(out);
	if (transparent_background) {
		imlib_image_set_has_alpha( 1 );
	}
	if (greenscreen) {
		imlib_context_set_color(0, 8, 0, 255);
	}else{
		if (invert_colors)
			imlib_context_set_color(0, 0, 0, (transparent_background)?0:255);
		else
			imlib_context_set_color(255, 255, 255, (transparent_background)?0:255);
	}
	imlib_image_fill_rectangle (0, 0, _max_used_linewidth, 8*_line);
	if (greenscreen) {
		imlib_context_set_color(20, 255, 40, 255);
	}else{
		if (invert_colors)
			imlib_context_set_color(255, 255, 255, 255);
		else
			imlib_context_set_color(0, 0, 0, 255);
	}
	for (int line=0; line<_line; line++) {
		for (int x=0; x<linewidth[line]; x++) {
			const char cpx=outbuffer[line][x];
			for (int bit=0; bit<8; bit++) {
				if (cpx&(128>>bit))
					imlib_image_draw_line(x, line*8+bit, x, line*8+bit, 0);
			}
		}
	}
	if (format)
		imlib_image_set_format(format);
	imlib_save_image(filename);
	return 0;
}

int main(int argc, char**argv) {
	char * filename = strdup("out.png");
	char * format= strdup("png");

	struct option opts[] = {
		{"format", 1, 0, 'f'},
		{"greenscreen",0,0,'g'},
		{"invert-colors",0,0,'i'},
		{"transparent",0,0,'t'},
		{"no-margins",0,0,'n'},
		{"output",1,0,'o'},
		{"help",0,0,'h'},
		{"8px",0,0,'8'},
		{"singlespace",0,0,'1'},
		{NULL, 0, 0, 0}
	};
	const char *opts_help[] = {
		"Output format - may be png, jpeg, bmp and others depending on your imlib installation",
		"Green screen output - different colors",
		"Invert output colors - ignored if --greenscreen is used",
		"transparent background (depends on image file format) - ignored if --greenscreen is used",
		"Omit margins when writing out image",
		"Output file - file to write output to. Writing to stdout is currently not supported",
		"Show this help", // help
		"Use 8px font instead of 5px",
		"1px spacing between 5x7px chars instead of 2",
		NULL,
	};
	unsigned int maxl=0;
	struct option *opt;
	char ** help;
	char * progname = NULL;
	progname = (char *)strrchr(argv[0], '/');
	progname = progname ? (progname + 1) : argv[0];
	bool filename_set=false, format_set=false;
	bool input_set=false;
	while(1) {
		int i = getopt_long(argc, argv, "+f:gitn81o:h", opts, NULL);
		if (i==-1)
			break;
		switch (i) {
			case 'n':
				omit_margin=true;
				break;
			case 'i':
				invert_colors=true;
				break;
			case 't':
				transparent_background=true;
				break;
			case 'g':
				greenscreen=true;
				break;
			case '1':
				singleSpace=true;
				break;
			case 'f':
				free(format);
				if (optarg==NULL) {
					fprintf(stderr, "Error: --format needs an argument!\n");
					return 1;
				}
				format = strdup(optarg);
				format_set=true;
				break;
			case '8':
				use8px=true;
				break;
			case 'o':
				free(filename);
				if (optarg==NULL) {
					fprintf(stderr, "Error: --output needs an argument!\n");
					return 1;
				}
				filename = strdup(optarg);
				filename_set=true;
				break;
			case 'h':
			default:
				fprintf(stderr, "usage: %s [OPTIONS] [<input file>]\nValid options:\n", progname);
				for (opt = opts; opt->name; opt++)
					if (strlen(opt->name) > maxl) maxl = strlen(opt->name);
				for (opt = opts, help = const_cast<char**>(opts_help); opt->name; opt++, help++) {
					unsigned int s;
					s = strlen(opt->name);
					fprintf(stderr, " -%c, --%s   ", opt->val, opt->name);
					for (;s < maxl;s++) fprintf(stderr, " ");
					fprintf(stderr, "%s", help[0]);
					if (opt->has_arg == 1)
						fprintf(stderr, " (needs argument)\n");
					else if (opt->has_arg == 2)
						fprintf(stderr, " (argument optional)\n");
					else
						fprintf(stderr, "\n");
				}
				if  (i == 'h')
					exit(EXIT_SUCCESS);
				else
					exit(EXIT_FAILURE);
				break;
		}
	}
	if (filename_set&&(!format_set)) {
		format=NULL;
	}
	/*if (!feof(stdin))
		inputfile=stdin;
	*/
//	if (!inputfile)
//		return 0;
	//get filename if any
	if (optind < argc) {
		char * file=argv[optind++];
		if (strlen(file)) {
			if (NULL==(inputfile=fopen(file,"r"))) { //fopen failed
				fprintf(stderr, "Error: could not open input file %s\n",file);
				return 1;
			}
			input_set=true;
		}
	}
	if (!inputfile)
		return 0;
	int ret=0;
	while(!feof(inputfile)) {
		int tmp=fgetc(inputfile);
		if (tmp==EOF)
			break;
		if ((ret=epson_parse(tmp))) {
			if (ret<0) {
				fprintf(stderr,"Interpreter critical error %i!\nTerminating\n",ret);
			}
			if (ret==1) { //image done
				break;
			}
			return -1;
		}
	}
	save_image(format, filename);
}
