#include <stdlib.h>
#include <inttypes.h>
#include <stdio.h>

#define PSF2_MAGIC0     0x72
#define PSF2_MAGIC1     0xb5
#define PSF2_MAGIC2     0x4a
#define PSF2_MAGIC3     0x86

typedef struct {
        unsigned char magic[4];
        uint32_t version;
        uint32_t headersize;    /* offset of bitmaps in file */
        uint32_t flags;
        uint32_t length;        /* number of glyphs */
        uint32_t charsize;      /* number of bytes for each character */
        uint32_t height, width; /* max dimensions of glyphs */
        /* charsize = height * ((width + 7) / 8) */
} psf2_header_t;

int charsize=0, fontw=0, fonth=0;

int loadfont(char * font) {
	psf2_header_t header;
	FILE* psf=fopen(font,"r");
	if (!psf) {
		return -1;
	}
	if (1!=fread(&header,sizeof(header),1,psf)) {
		return -1;
	}
	//fseek(psf,header.headersize,SEEK_SET);
	font=(char*)calloc(header.charsize,header.length);
	if (!font) {
		printf("Error alloc'ing memory for font\n");
		return -1;
	}
	fread(font,header.charsize,header.length,psf);
	fclose(psf);
	charsize=header.charsize;
	fontw=header.width;
	fonth=header.height;
	return 0;
}

unsigned char * glyphToRaster(char * font, int c, int * bytes) {
	int cpitch=(fontw+7)/8;
	char cdata=0;
	char* data;
	int pitch=(fonth+7)/8;
	unsigned char * buffer=(unsigned char*)calloc((*bytes)=pitch*fontw,1);
	for(int cy=0;cy<fonth;cy++) {
		data=&font[c*charsize+cy*cpitch];
		cdata=*data++;
		for ( int cx=0;cx<fontw;cx++) {
			if (cdata&0x80)
				buffer[cx*pitch+(cy/8)]|=(1>>(cy&0x07));
			cdata<<=1;
			if(((cx&7)==7)&&(fontw>8))
				cdata=*(data++);
		}
	}
	return buffer;
}

int main (int argc, char ** argv) {
	if (argc!=2) {
		fprintf(stderr,"USAGE: %s <font file>\n",argv[0]);
		return 1;
	}
	if (loadfont(argv[2])) {
		fprintf(stderr,"invalid font file\n",argv[0]);
		return 1;
	}
}